"""
Contains the configuration for the ixoncdkingress webserver
"""
import enum
import json
import os
import sys
from collections.abc import Mapping
from dataclasses import dataclass
from logging import Formatter, Logger, StreamHandler, getLogger
from typing import Any

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from load_environ_typed import LoaderMap, load
from load_environ_typed.loaders import load_utf8_file_from_path


class InvalidConfigException(Exception):
    """
    Exception for invalid values in the config
    """

class WsgiProvider(enum.Enum):
    """
    Provider to use to turn a WSGI application into a (HTTP) server
    """
    WSGIREF = 'wsgiref'
    BJOERN = 'bjoern'
    WAITRESS = 'waitress'
    GUNICORN = 'gunicorn'
    CHERRYPY = 'cherrypy'
    MEINHELD = 'meinheld'

@dataclass
class Config:
    """
    Configuration for the webserver
    """
    production_mode: bool

    document_db_port: int

    http_server_bind: str
    http_server_port: int
    wsgi_provider: WsgiProvider

    logger_log_level: str
    logger_format: str

    cbc_path: str
    api_client_base_url: str

    cbc_signature_public_keys: list[str]

    context_values_path: str
    context_config_path: str

    _logger: Logger | None = None
    _signature_public_keys: list[ed25519.Ed25519PublicKey] | None = None

    @classmethod
    def from_environ(cls, environ: Mapping[str, str]) -> 'Config':
        """
        Deprecated, use get_config() instead.

        Parses an environment dictionary and generates a config based on it
        """
        return get_config(environ)

    def get_logger(self) -> Logger:
        """
        Returns a logger
        """
        if self._logger is None:
            logger = getLogger('ixoncdkingress')
            logger.setLevel(self.logger_log_level)

            handler = StreamHandler(sys.stdout)
            handler.setFormatter(Formatter(self.logger_format))
            logger.addHandler(handler)

            self._logger = logger

        return self._logger

    def get_signature_public_keys(self) -> list[ed25519.Ed25519PublicKey]:
        """
        Returns a cached instance of signature public keys
        """
        if self._signature_public_keys is None:
            keys: list[ed25519.Ed25519PublicKey] = []

            for index, pem in enumerate(self.cbc_signature_public_keys):
                key = load_pem_public_key(pem.encode('ASCII'), None)
                assert isinstance(key, ed25519.Ed25519PublicKey), \
                    f'signature_public_key {index + 1} is not a valid Ed25519 private key'
                keys.append(key)

            self._signature_public_keys = keys
        return self._signature_public_keys

def load_str_file(
        file_path: str | None, default: str
    ) -> str:
    """
    Loads a string from a file, returns default if no path is given
    """
    if not file_path:
        return default

    try:
        with open(file_path, encoding='utf-8') as file:
            string = file.read().rstrip()
            if not string:
                return default
        return string
    except OSError:
        return default

def load_json_file(
        file_path: str,
    ) -> dict[str, Any]:
    """
    Loads a Dict from a JSON file, converting all numeric values to string. Returns default if no
    path is given, file_path does not exist, or file_path does not contain a JSON file.
    """
    with open(file_path, encoding='utf-8') as file:
        result = json.load(file)

    if not isinstance(result, dict):
        raise ValueError('must be a JSON object')

    return result

def load_context_values(config: Config) -> dict[str, Any]:
    """
    Loads context values from a config file. First checks the file specified by
    CONTEXT_VALUES_PATH environment variable, then loads it as JSON into a dict.
    If that file does not exist, checks the file specified by
    CONTEXT_CONFIG_PATH environment variable, then loads it as JSON into a dict
    and returns only the 'values' field. It also writes the values as JSON to a
    file at CONTEXT_VALUES_PATH for future invocations.
    """
    logger = config.get_logger()
    values_path = config.context_values_path
    config_path = config.context_config_path

    if os.path.exists(values_path):
        logger.info("Loading context values from %s", values_path)
        if os.path.exists(config_path):
            logger.info("The file %s is unused and can be safely removed", config_path)

        return load_json_file(values_path)

    if os.path.exists(config_path):
        logger.info("Loading context config from %s", config_path)

        context_config = load_json_file(config_path)
        context_values = context_config.get('values')

        if isinstance(context_values, dict):
            logger.warning("The file %s is deprecated", config_path)
            logger.warning(
                "Contents of 'values' field from %s copied to %s.",
                config_path,
                values_path,
            )
            logger.warning(
                "The file %s will no longer be used and can be safely removed.",
                config_path
            )
            with open(values_path, 'w') as f:
                json.dump(context_values, f)

            return context_values

    return {}

def load_comma_separated_strings(raw: str) -> list[str]:
    return raw.split(',') if raw != '' else []

CONFIG_DEFAULTS = {
    'API_CLIENT_BASE_URL': 'https://api.ayayot.com/',
    'CBC_PATH': '/user_scripts',
    'CBC_SIGNATURE_PUBLIC_KEYS': '',
    'CONTEXT_VALUES_PATH': './context_values.json',
    'CONTEXT_CONFIG_PATH': './context_config.json',
    'DOCUMENT_DB_PORT': '27017',
    'HTTP_SERVER_BIND': '127.0.0.1',
    'HTTP_SERVER_PORT': '8020',
    'LOGGER_FORMAT': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    'LOGGER_LOG_LEVEL': 'INFO',
    'PRODUCTION_MODE': 'False',
    'WSGI_PROVIDER': 'wsgiref',
}

CONFIG_LOADERS: LoaderMap = {
    'cbc_signature_public_keys': load_comma_separated_strings,
    'version': load_utf8_file_from_path,
    'wsgi_provider': WsgiProvider,
}

def get_config(environ: Mapping[str, str] | None = None) -> Config:
    return load(Config, environ=environ, defaults=CONFIG_DEFAULTS, loaders=CONFIG_LOADERS)



"""
Context.
"""
from inspect import _empty, signature
from typing import Any

from pymongo import MongoClient

from ixoncdkingress.function.api_client import ApiClient
from ixoncdkingress.function.document_db_client import DocumentDBClient, DocumentType
from ixoncdkingress.function.time_series_client import TimeSeriesClient


class FunctionResource:
    """
    Describes an IXAPI resource.
    """
    public_id: str
    name: str
    custom_properties: dict[str, Any]
    permissions: set[str] | None

    def __init__(
            self,
            public_id: str,
            name: str,
            custom_properties: dict[str, Any],
            permissions: set[str] | None
        ) -> None:
        self.public_id = public_id
        self.name = name
        self.custom_properties = custom_properties
        self.permissions = permissions

    def __repr__(self) -> str:
        return (
            '<FunctionResource'
            f' public_id={self.public_id},'
            f' name={self.name},'
            f' custom_properties={self.custom_properties!r},'
            f' permissions={self.permissions!r},'
            f'>'
        )

class FunctionContext:
    """
    The context for a Cloud Function.
    """
    config: dict[str, str]
    api_client: ApiClient
    user: FunctionResource | None
    company: FunctionResource | None
    asset: FunctionResource | None
    agent: FunctionResource | None

    _mongo_client: MongoClient[DocumentType] | None
    _ddb_collection_name: str

    @property
    def document_db_client(self) -> DocumentDBClient | None:
        """
        The client for accessing the default collection of the Document DB.

        This may be None when running the Cloud Function in development mode,
        and using the local debug form without specifying a company.
        """
        if not self._mongo_client or not self.company:
            return None

        return DocumentDBClient(
            self._mongo_client[self.company.public_id][self._ddb_collection_name]
        )

    @property
    def time_series_client(self) -> TimeSeriesClient | None:
        """
        The client for accessing the time series collection of the Cloud Function.

        This client is experimental.

        This may be None when running the Cloud Function in development mode,
        and using the local debug form without specifying a company.
        """
        if not self._mongo_client or not self.company:
            return None

        return TimeSeriesClient(
            self._mongo_client[self.company.public_id][f"{self._ddb_collection_name}_ts"]
        )

    @property
    def agent_or_asset(self) -> FunctionResource:
        """
        Return either an Agent or an Asset resource, depending on what's available. If both are
        available in the context, returns the Asset resource.
        """
        if self.asset:
            return self.asset

        assert self.agent
        return self.agent

    def __init__(
            self,
            config: dict[str, str],
            api_client: ApiClient,
            mongo_client: MongoClient[DocumentType] | None,
            user: FunctionResource | None,
            company: FunctionResource | None,
            asset: FunctionResource | None,
            agent: FunctionResource | None,
            document_db_collection_name: str,
        ) -> None:

        self.config = config
        self.api_client = api_client
        self.user = user
        self.company = company
        self.asset = asset
        self.agent = agent

        self._mongo_client = mongo_client
        self._ddb_collection_name = document_db_collection_name

    def __repr__(self) -> str:
        return (
            f'<FunctionContext'
            f' config={self.config!r},'
            f' api_client={self.api_client!r},'
            f' document_db_client={self.document_db_client!r},'
            f' user={self.user!r},'
            f' company={self.company!r},'
            f' asset={self.asset!r},'
            f' agent={self.agent!r},'
            f'>'
        )

    @staticmethod
    def expose(function: Any) -> Any:
        """
        Decorator to mark a function as an exposed endpoint.
        """
        sig = signature(function, eval_str=True)

        if not sig.parameters:
            raise Exception('Function has no argument for FunctionContext')

        # If the first function argument has a type annotation it should be of FunctionContext
        context_param = sig.parameters[next(iter(sig.parameters))]
        if (context_param.annotation is not _empty
                and context_param.annotation is not FunctionContext):
            raise Exception('First function parameter should be of type FunctionContext')

        function.exposed = True

        return function



"""
Module containing type aliases and enums
"""
import enum

FunctionLocation = tuple[str, str]
"""
Location of the function. First element specifies the Python module path, second element the
Python function name.
"""

FunctionArguments = dict[str, str]
"""Arguments to the function."""

class ContentType(enum.Enum):
    """
    HTTP content types which are used in the Content-Type headers
    """
    FORM = 'application/x-www-form-urlencoded'
    HTML = 'text/html'
    JSON = 'application/json'

class RequestMethod(enum.Enum):
    """
    Enum with valid HTTP methods
    """
    GET = 'GET'
    POST = 'POST'
    PATCH = 'PATCH'
    DELETE = 'DELETE'
    PUT = 'PUT'
    OPTIONS = 'OPTIONS'

class ResponseCode(enum.Enum):
    """
    HTTP response codes to return
    """
    OK = 200
    NO_CONTENT = 204
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    METHOD_NOT_ALLOWED = 405
    INTERNAL_ERROR = 500
    NOT_AVAILABLE = 503

    def status_line(self) -> str:
        """Produces a HTTP status line: <status-code> <message>"""
        return f"{self.value} {self.name.replace('_', ' ')}"



import warnings

warnings.warn(
    "ixoncdkingress.cbc had been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.api_client import ApiClient # noqa: E402, I001

__all__ = ["ApiClient"]



"""
Types that can be used when implementing
the Object Storage interface.

Compatible with ayayot.objectstorage_v1
"""

import enum
from typing import Literal, TypedDict


class ResourceType(str, enum.Enum):
    """
    A type of resource for which a path is returned
    """
    ASSET = 'Asset'
    AGENT = 'Agent'

class PathData(TypedDict, total=True):
    """
    Response data for a single path
    """
    path: str

class PathMapping(TypedDict, total=True):
    """
    A mapping between a certain resource and a path,
    as returned by authorize_list
    """
    publicId: str
    path: str
    type: ResourceType

class PathResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    Object Storage authorize_upload, authorize_download &
    authorize_delete APIs
    """
    result: Literal['success']
    data: PathData

class ListPathResponse(TypedDict, total=True):
    """
    The expected type for a response for the
    Object Storage authorize_list API
    """
    result: Literal['success']
    data: list[PathMapping]



"""
HTML form generator
"""
from typing import TYPE_CHECKING
from urllib.parse import parse_qs

from ixoncdkingress.function.api_client import ApiClient
from ixoncdkingress.function.context import FunctionContext, FunctionResource
from ixoncdkingress.types import FunctionArguments, FunctionLocation
from ixoncdkingress.webserver.config import Config
from ixoncdkingress.webserver.utils import (
    create_mongo_client,
    get_document_db_collection_name,
    parse_function_location,
)

try:
    from dominate.document import document
    from dominate.tags import (
        div,
        form,
        h1,
        h3,
        input_,
        label,
        table,
        tbody,
        td,
        textarea,
        th,
        thead,
        tr,
    )

    HAS_DOMINATE = True # pragma: no cover
except ImportError: # pragma: no cover
    HAS_DOMINATE = False

if TYPE_CHECKING:
    from dominate import tags

def generate_form(pre_fill: dict[str, str], form_output: str | None = None) -> str:
    """
    Generates and renders the CDK Ingress HTML form
    """
    return _generate_form_document(pre_fill, form_output).render()

def _generate_form_document(
        pre_fill: dict[str, str],
        form_output: str | None = None
    ) -> 'document':
    """
    Generates the CDK Ingress HTML form document
    """
    if not HAS_DOMINATE:
        raise Exception('Requires dominate package. pip install dominate')

    doc = document(title='IXON CDK Ingress')

    with doc:
        h1('IXON CDK Ingress')
        with form(method='post'):
            with div(class_name='function'):
                label('Function', _for='function')
                input_(type='text', id='function', name='function', size='32', required='true',
                       value=pre_fill.get('function'))
            with div(class_name='arguments'):
                h3('Function input')
                _custom_properties(pre_fill, prefix='function_args__')

            with div(class_name='resource'):
                h3('User')
                _resource(pre_fill, prefix='user__', show_permission=False)
            with div(class_name='resource'):
                h3('Company')
                _resource(pre_fill, prefix='company__', show_permission=True)
            with div(class_name='resource'):
                h3('Asset')
                _resource(pre_fill, prefix='asset__', show_permission=True)
            with div(class_name='resource'):
                h3('Agent')
                _resource(pre_fill, prefix='agent__', show_permission=True)

            input_(type='submit', value='Run function')
            input_(type='reset', value='Clear form')

        h3('Function output')
        textarea(form_output or 'Function output goes here', rows='10', cols='80')

    return doc

def _custom_properties(pre_fill: dict[str, str], prefix: str) -> 'tags.html_tag':
    """
    Generates a form fragment for custom properties.
    """
    table_tag = table()

    with table_tag:
        with thead():
            th('Key')
            th('Value')
        with tbody():
            for i in range(9):
                with tr():
                    td(input_(type='text', name=f'{prefix}key{i}', id=f'{prefix}key{i}',
                              size='16', value=pre_fill.get(f'{prefix}key{i}')))
                    td(input_(type='text', name=f'{prefix}value{i}', id=f'{prefix}value{i}',
                              size='32', value=pre_fill.get(f'{prefix}value{i}')))

    return table_tag

def _permissions(pre_fill: dict[str, str], prefix: str) -> 'tags.html_tag':
    """
    Generates a form fragment for permissions.
    """
    table_tag = table()

    with table_tag, tbody():
        for i in range(9):
            with tr():
                td(input_(type='text', name=f'{prefix}value{i}', id=f'{prefix}value{i}',
                          size='32', value=pre_fill.get(f'{prefix}value{i}')))

    return table_tag

def _resource(pre_fill: dict[str, str], prefix: str, show_permission: bool) -> 'tags.html_tag':
    """
    Generates a form fragment for a context resource.
    """
    table_tag = table()

    with table_tag, tbody():
        with tr():
            td(label('PublicID', _for=f'{prefix}public_id'))
            td(input_(type='text', id=f'{prefix}public_id', name=f'{prefix}public_id',
                      size='12', value=pre_fill.get(f'{prefix}public_id')))
        with tr():
            td(label('Name', _for=f'{prefix}name'))
            td(input_(type='text', id=f'{prefix}name', name=f'{prefix}name', size='32',
                      value=pre_fill.get(f'{prefix}name')))
        with tr():
            td('Custom properties')
            td(_custom_properties(pre_fill, f'{prefix}custom_properties__'))

        if show_permission:
            td('Permissions')
            td(_permissions(pre_fill, f'{prefix}permissions__'))

    return table_tag

def parse_form_input(
        config: Config,
        context_config: dict[str, str],
        body: str) -> tuple[FunctionContext, FunctionLocation, FunctionArguments, str]:
    """
    Parses an application/x-www-form-urlencoded request body string into a context, function
    location and function arguments.
    """

    in_put = parse_qs(body)

    context = FunctionContext(
        context_config,
        api_client=ApiClient(config.api_client_base_url),
        mongo_client=create_mongo_client(config, {}),
        user=_parse_resource('user__', in_put, False),
        company=_parse_resource('company__', in_put, True),
        asset=_parse_resource('asset__', in_put, True),
        agent=_parse_resource('agent__', in_put, True),
        document_db_collection_name=get_document_db_collection_name()
    )

    function_location = parse_function_location(in_put.get('function', [''])[0])
    function_arguments = _parse_function_args(in_put)

    return context, function_location, function_arguments, ''

def _parse_function_args(in_put: dict[str, list[str]]) -> FunctionArguments:
    function_arguments = {}
    for in_put_key, in_put_value in in_put.items():
        if not in_put_key.startswith('function_args__key'):
            continue

        key_argument_value = f'function_args__value{in_put_key[18:]}'
        if key_argument_value not in in_put:
            continue

        function_arguments[in_put_value[0]] = in_put[key_argument_value][0]

    return function_arguments

def _parse_resource(
        prefix: str,
        in_put: dict[str, list[str]],
        has_permissions: bool
    ) -> FunctionResource | None:
    """
    Reads a resource description from the in_put and returns it as a FunctionResource.
    """
    props = {}
    permissions: set[str] | None = set() if has_permissions else None
    if not in_put.get(f'{prefix}public_id') or not in_put.get(f'{prefix}name'):
        return None

    custom_property_prefix = f'{prefix}custom_properties__'
    permission_prefix = f'{prefix}permissions__value'
    for in_put_key, in_put_value in in_put.items():

        # Process custom property
        if in_put_key.startswith(f'{custom_property_prefix}key'):
            custom_property_index = in_put_key[len(f'{custom_property_prefix}key'):]
            key_custom_property_value = f'{custom_property_prefix}value{custom_property_index}'
            if key_custom_property_value not in in_put:
                continue

            props[in_put_value[0]] = in_put[key_custom_property_value][0]
            continue

        # Process permission
        if has_permissions and in_put_key.startswith(permission_prefix):
            assert isinstance(permissions, set) # type hint
            permissions.add(in_put_value[0])

    return FunctionResource(in_put[f'{prefix}public_id'][0],
                            in_put[f'{prefix}name'][0],
                            props,
                            permissions)



"""
Functions for finding and calling Cloud Functions.
"""
import importlib
import sys
from typing import Any

from pymongo.errors import OperationFailure

from ixoncdkingress.function.context import FunctionContext
from ixoncdkingress.types import FunctionArguments, FunctionLocation
from ixoncdkingress.utils import handle_exception
from ixoncdkingress.webserver.config import Config
from ixoncdkingress.webserver.response import Response


def call_function(
        config: Config, context: FunctionContext,
        function_location: FunctionLocation,
        function_kwargs: FunctionArguments,
        response: Response
    ) -> Any:
    """
    Finds, loads and calls the function specified in the body. The content_type specifies the
    format of the body.
    """

    # Get the specified module
    sys.path.insert(0, config.cbc_path)
    module = importlib.import_module(function_location[0])
    if config.production_mode is False:
        try:
            module = importlib.reload(module)
        except ImportError as error:
            handle_exception(error, response)
            return None
    del sys.path[0]

    # Get the specified function
    if hasattr(module, function_location[1]):
        function = getattr(module, function_location[1])

        # Check if it's exposed
        if getattr(function, 'exposed', False):
            try:
                return function(context, **function_kwargs)
            except OperationFailure as ex:
                if ex.details and ex.details.get('codeName'):
                    return f'Document DB operation failure: {ex.details["codeName"]}'

                return 'Document DB operation failure'


    return f"Not found, function '{'.'.join(function_location)}' does not exist or is not exposed"
